/*
 * Copyright 2009 Maarten Maathuis
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
 * OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/**
 * This file serves to process all damage to the frontbuffer pixmap.
 * This is needed until we can/have:
 * - Tiled frontbufffer support.
 * - Working wfb support.
 * - Widespread wfb exa support.
 */

#include "nv_include.h"
#include "damagestr.h"

/* When driver allocated pixmaps are used we can easily fold this back into exa code. */

static void
nv50_shadow_damage_blit_state_emit(struct nouveau_channel *chan)
{
	ScrnInfoPtr pScrn = chan->user_private;
	NVPtr pNv = NVPTR(pScrn);
	PixmapPtr ppix = pNv->pspix;
	struct nouveau_bo *bo = nouveau_pixmap_bo(ppix);
	struct nouveau_grobj *eng2d = pNv->Nv2D;
	unsigned delta = nouveau_pixmap_offset(ppix), fmt;

	WAIT_RING (chan, 27 + 13);

	switch (ppix->drawable.depth) {
	case 8 : fmt = NV50_2D_SRC_FORMAT_8BPP; break;
	case 15: fmt = NV50_2D_SRC_FORMAT_15BPP; break;
	case 16: fmt = NV50_2D_SRC_FORMAT_16BPP; break;
	case 24: fmt = NV50_2D_SRC_FORMAT_24BPP; break;
	case 32: fmt = NV50_2D_SRC_FORMAT_32BPP; break;
	default:
		 xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
			    "Unknown surface format for bpp=%d\n",
			    ppix->drawable.depth);
		 return;
	}

	/* tiled source */
	BEGIN_RING(chan, eng2d, NV50_2D_SRC_FORMAT, 5);
	OUT_RING  (chan, fmt);
	OUT_RING  (chan, 0);
	OUT_RING  (chan, 0);
	OUT_RING  (chan, 1);
	OUT_RING  (chan, 0);

	BEGIN_RING(chan, eng2d, NV50_2D_SRC_WIDTH, 4);
	OUT_RING  (chan, ppix->drawable.width);
	OUT_RING  (chan, ppix->drawable.height);
	OUT_RELOCh(chan, bo, delta, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);
	OUT_RELOCl(chan, bo, delta, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);

	/* untiled destination */
	BEGIN_RING(chan, eng2d, NV50_2D_DST_FORMAT, 2);
	OUT_RING  (chan, fmt);
	OUT_RING  (chan, 1);
	BEGIN_RING(chan, eng2d, NV50_2D_DST_PITCH, 5);
	OUT_RING  (chan, NOUVEAU_ALIGN(pScrn->virtualX, 64) *
			 (pScrn->bitsPerPixel >> 3));
	OUT_RING  (chan, ppix->drawable.width);
	OUT_RING  (chan, ppix->drawable.height);
	OUT_RELOCh(chan, pNv->scanout, 0, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);
	OUT_RELOCl(chan, pNv->scanout, 0, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);

	BEGIN_RING(chan, eng2d, NV50_2D_CLIP_X, 4);
	OUT_RING  (chan, 0);
	OUT_RING  (chan, 0);
	OUT_RING  (chan, ppix->drawable.width);
	OUT_RING  (chan, ppix->drawable.height);
	BEGIN_RING(chan, eng2d, NV50_2D_OPERATION, 1);
	OUT_RING  (chan, NV50_2D_OPERATION_SRCCOPY);
}

static void
nv50_shadow_damage_blit(PixmapPtr ppix, RegionPtr pRegion)
{
	ScrnInfoPtr pScrn = xf86Screens[ppix->drawable.pScreen->myNum];
	NVPtr pNv = NVPTR(pScrn);
	struct nouveau_channel *chan = pNv->chan;
	struct nouveau_grobj *eng2d = pNv->Nv2D;
	BoxPtr pbox;
	int nbox;

	pbox = REGION_RECTS(pRegion);
	nbox = REGION_NUM_RECTS(pRegion);
	if (!nbox)
		return;

	pNv->pspix = ppix;
	chan->flush_notify = nv50_shadow_damage_blit_state_emit;
	chan->flush_notify(chan);
	while (nbox--) {
		WAIT_RING (chan, 13);
		BEGIN_RING(chan, eng2d, NV50_2D_BLIT_DST_X, 12);
		OUT_RING  (chan, pbox->x1);
		OUT_RING  (chan, pbox->y1);
		OUT_RING  (chan, pbox->x2 - pbox->x1);
		OUT_RING  (chan, pbox->y2 - pbox->y1);
		OUT_RING  (chan, 0);
		OUT_RING  (chan, 1);
		OUT_RING  (chan, 0);
		OUT_RING  (chan, 1);
		OUT_RING  (chan, 0);
		OUT_RING  (chan, pbox->x1);
		OUT_RING  (chan, 0);
		OUT_RING  (chan, pbox->y1);

		pbox++;
	}
	chan->flush_notify = NULL;
}

/* For frontbuffer fallbacks. */
static void
nv50_shadow_damage_blit_back_state_emit(struct nouveau_channel *chan)
{
	ScrnInfoPtr pScrn = chan->user_private;
	NVPtr pNv = NVPTR(pScrn);
	PixmapPtr ppix = pNv->pdpix;
	struct nouveau_grobj *eng2d = pNv->Nv2D;
	struct nouveau_bo *bo = nouveau_pixmap_bo(ppix);
	unsigned delta = nouveau_pixmap_offset(ppix), fmt;

	WAIT_RING (chan, 27 + 13);

	switch (ppix->drawable.depth) {
	case 8 : fmt = NV50_2D_SRC_FORMAT_8BPP; break;
	case 15: fmt = NV50_2D_SRC_FORMAT_15BPP; break;
	case 16: fmt = NV50_2D_SRC_FORMAT_16BPP; break;
	case 24: fmt = NV50_2D_SRC_FORMAT_24BPP; break;
	case 32: fmt = NV50_2D_SRC_FORMAT_32BPP; break;
	default:
		 xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
			    "Unknown surface format for bpp=%d\n",
			    ppix->drawable.depth);
		 return;
	}

	/* untiled source */
	BEGIN_RING(chan, eng2d, NV50_2D_SRC_FORMAT, 2);
	OUT_RING  (chan, fmt);
	OUT_RING  (chan, 1);
	BEGIN_RING(chan, eng2d, NV50_2D_SRC_PITCH, 5);
	OUT_RING  (chan, NOUVEAU_ALIGN(pScrn->virtualX, 64) *
			 (pScrn->bitsPerPixel >> 3));
	OUT_RING  (chan, ppix->drawable.width);
	OUT_RING  (chan, ppix->drawable.height);
	OUT_RELOCh(chan, pNv->scanout, 0, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);
	OUT_RELOCl(chan, pNv->scanout, 0, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);

	/* tiled destination */
	BEGIN_RING(chan, eng2d, NV50_2D_DST_FORMAT, 5);
	OUT_RING  (chan, fmt);
	OUT_RING  (chan, 0);
	OUT_RING  (chan, 0);
	OUT_RING  (chan, 1);
	OUT_RING  (chan, 0);

	BEGIN_RING(chan, eng2d, NV50_2D_DST_WIDTH, 4);
	OUT_RING  (chan, ppix->drawable.width);
	OUT_RING  (chan, ppix->drawable.height);
	OUT_RELOCh(chan, bo, delta, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);
	OUT_RELOCl(chan, bo, delta, NOUVEAU_BO_VRAM | NOUVEAU_BO_WR);

	BEGIN_RING(chan, eng2d, NV50_2D_CLIP_X, 4);
	OUT_RING  (chan, 0);
	OUT_RING  (chan, 0);
	OUT_RING  (chan, ppix->drawable.width);
	OUT_RING  (chan, ppix->drawable.height);
	BEGIN_RING(chan, eng2d, NV50_2D_OPERATION, 1);
	OUT_RING  (chan, NV50_2D_OPERATION_SRCCOPY);
}

static void
nv50_shadow_damage_blit_back(PixmapPtr ppix, RegionPtr pRegion)
{
	ScrnInfoPtr pScrn = xf86Screens[ppix->drawable.pScreen->myNum];
	NVPtr pNv = NVPTR(pScrn);
	struct nouveau_channel *chan = pNv->chan;
	struct nouveau_grobj *eng2d = pNv->Nv2D;
	BoxPtr pbox;
	int nbox;

	pbox = REGION_RECTS(pRegion);
	nbox = REGION_NUM_RECTS(pRegion);
	if (!nbox)
		return;

	pNv->pdpix = ppix;
	chan->flush_notify = nv50_shadow_damage_blit_back_state_emit;
	chan->flush_notify(chan);
	while (nbox--) {
		WAIT_RING (chan, 13);
		BEGIN_RING(chan, eng2d, NV50_2D_BLIT_DST_X, 12);
		OUT_RING  (chan, pbox->x1);
		OUT_RING  (chan, pbox->y1);
		OUT_RING  (chan, pbox->x2 - pbox->x1);
		OUT_RING  (chan, pbox->y2 - pbox->y1);
		OUT_RING  (chan, 0);
		OUT_RING  (chan, 1);
		OUT_RING  (chan, 0);
		OUT_RING  (chan, 1);
		OUT_RING  (chan, 0);
		OUT_RING  (chan, pbox->x1);
		OUT_RING  (chan, 0);
		OUT_RING  (chan, pbox->y1);

		pbox++;
	}
	chan->flush_notify = NULL;
}

static void
nv50_shadow_damage_report(DamagePtr pDamage, RegionPtr pRegion, void *closure)
{
	PixmapPtr ppix = closure;

	nv50_shadow_damage_blit(ppix, pRegion);
}

void
nv50_shadow_damage_frontbuffer_fallback(ScrnInfoPtr pScrn)
{
	NVPtr pNv = NVPTR(pScrn);
	ScreenPtr pScreen = pScrn->pScreen;
	PixmapPtr ppix = NULL;
	DamagePtr pDamage = pNv->screen_damage;

	if (pNv->Architecture < NV_ARCH_50)
		return;

	ppix = pScreen->GetScreenPixmap(pScreen);
	if (!ppix) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
			"No screen pixmap.\n");
		return;
	}

	/* Pending damage reflects the rendering currently being done. */
	/* When exa calls finish access, damage hasn't flushed it yet. */
	nv50_shadow_damage_blit_back(ppix, &pDamage->pendingDamage);
}

static void
nv50_shadow_damage_destroy(DamagePtr pDamage, void *closure)
{
	PixmapPtr ppix = closure;
	ScrnInfoPtr pScrn = xf86Screens[ppix->drawable.pScreen->myNum];
	NVPtr pNv = NVPTR(pScrn);

	pNv->screen_damage = NULL;
}

bool
nv50_shadow_damage_create(ScrnInfoPtr pScrn)
{
	NVPtr pNv = NVPTR(pScrn);
	ScreenPtr pScreen = pScrn->pScreen;
	PixmapPtr ppix = NULL;

	if (pNv->Architecture < NV_ARCH_50)
		return false;

	ppix = pScreen->GetScreenPixmap(pScreen);
	if (!ppix) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
			"No screen pixmap.\n");
		return false;
	}

	pNv->screen_damage = DamageCreate(nv50_shadow_damage_report, nv50_shadow_damage_destroy, DamageReportRawRegion, true, pScreen, ppix);
	if (!pNv->screen_damage) {
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
			"No screen damage\n");
		return false;
	}

	/* We want the notification after submission. */
	DamageSetReportAfterOp(pNv->screen_damage, true);

	DamageRegister(&ppix->drawable, pNv->screen_damage);

	return true;
}
